import dolfin
import numpy as np
import ufl

from . import assemble
from . import Constant
from . import Function
from . import FunctionAssigner
from . import numpy_mpi
from . import project
from .utils import DOLFIN_VERSION_MAJOR
from .utils import getLogger

logger = getLogger(__name__)


def map_vector_field(f0, new_mesh, u=None, name="fiber", normalize=True):
    """
    Map a vector field (f0) onto a new mesh (new_mesh) where the new mesh
    can be a moved version of the original one according to some
    displacement (u). In that case we will just a Piola transform to
    map the vector field.
    """
    representation = dolfin.parameters["form_compiler"]["representation"]
    if DOLFIN_VERSION_MAJOR > 2016:
        dolfin.parameters["form_compiler"]["representation"] = "quadrature"

    dolfin.parameters["form_compiler"]["quadrature_degree"] = 4

    ufl_elem = f0.function_space().ufl_element()
    f0_new = Function(dolfin.FunctionSpace(new_mesh, ufl_elem))

    if u is not None:
        f0_mesh = f0.function_space().mesh()
        u_elm = u.function_space().ufl_element()
        V = dolfin.FunctionSpace(f0_mesh, u_elm)
        u0 = Function(V)
        # arr = numpy_mpi.gather_vector(u.vector())
        # numpy_mpi.assign_to_vector(u0.vector(), arr)
        u0.vector()[:] = u.vector()
        from .kinematics import DeformationGradient

        F = DeformationGradient(u0)

        f0_updated = project(F * f0, f0.function_space())

        if normalize:
            f0_updated = normalize_vector_field(f0_updated)

        f0_new.vector()[:] = f0_updated.vector()
        # f0_arr = numpy_mpi.gather_vector(f0_updated.vector())
        # numpy_mpi.assign_to_vector(f0_new.vector(), f0_arr)

    else:
        # f0_arr = numpy_mpi.gather_vector(f0.vector())
        # numpy_mpi.assign_to_vector(f0_new.vector(), f0_arr)
        f0_new.vector()[:] = f0.vector()

    if DOLFIN_VERSION_MAJOR > 2016:
        dolfin.parameters["form_compiler"]["representation"] = representation

    return f0_new


def update_function(mesh, f):
    """Given a function :math:`f` defined on some domain,
    update the function so that it now is defined on the domain
    given in the mesh
    """

    f_new = Function(dolfin.FunctionSpace(mesh, f.ufl_element()))
    numpy_mpi.assign_to_vector(f_new.vector(), numpy_mpi.gather_vector(f.vector()))
    return f_new


def normalize_vector_field(u):
    """Given a vector field, return a vector field with an L2 norm equal to 1.0"""
    dim = len(u)
    S = u.function_space().sub(0).collapse()

    components = vectorfield_to_components(u, S, dim)

    normarray = np.sqrt(sum(components[i].vector().norm("l2") ** 2 for i in range(dim)))

    for comp in components:
        comp.vector()[:] = comp.vector() / normarray

    assigners = [FunctionAssigner(u.function_space().sub(i), S) for i in range(dim)]
    for i, comp, assigner in zip(range(dim), components, assigners):
        assigner.assign(u.split()[i], comp)

    return u


def vectorfield_to_components(u, S, dim):
    components = [Function(S) for i in range(dim)]
    assigners = [FunctionAssigner(S, u.function_space().sub(i)) for i in range(dim)]
    for i, comp, assigner in zip(range(dim), components, assigners):
        assigner.assign(comp, u.split()[i])

    return components


def get_pressure(problem):
    """Returns p_lv (and p_rv if BiV mesh)"""

    plv = [p.traction for p in problem.bcs.neumann if p.name == "lv"]
    prv = [p.traction for p in problem.bcs.neumann if p.name == "rv"]

    assert len(plv) > 0, "Problem has no Neumann BC for LV endo"
    pressure = [plv[0]]
    if prv:
        pressure.append(prv[0])
        return tuple(pressure)
    else:
        return pressure[0]


def compute_meshvolume(domain=None, dx=dolfin.dx, subdomain_id=None):
    return Constant(
        assemble(
            Constant(1.0) * dx(domain=domain, subdomain_id=subdomain_id),
        ),
    )


def get_cavity_volume(geometry, unload=False, chamber="lv", u=None, xshift=0.0):
    if unload:
        mesh = geometry.original_geometry
        ffun = dolfin.MeshFunction("size_t", mesh, 2, mesh.domains())
    else:
        mesh = geometry.mesh
        ffun = geometry.ffun

    if chamber == "lv":
        if "ENDO_LV" in geometry.markers:
            endo_marker = geometry.markers["ENDO_LV"]
        else:
            endo_marker = geometry.markers["ENDO"]

    else:
        endo_marker = geometry.markers["ENDO_RV"]

    if hasattr(endo_marker, "__len__"):
        endo_marker = endo_marker[0]

    ds = dolfin.Measure("exterior_facet", subdomain_data=ffun, domain=mesh)(endo_marker)

    vol_form = get_cavity_volume_form(geometry.mesh, u, xshift)
    return assemble(vol_form * ds)


def get_cavity_volume_form(mesh, u=None, xshift=0.0):
    from . import kinematics

    shift = Constant((xshift, 0.0, 0.0))
    X = dolfin.SpatialCoordinate(mesh) - shift
    N = dolfin.FacetNormal(mesh)

    if u is None:
        vol_form = (-1.0 / 3.0) * dolfin.dot(X, N)
    else:
        F = kinematics.DeformationGradient(u)
        J = kinematics.Jacobian(F)
        vol_form = (-1.0 / 3.0) * dolfin.dot(X + u, J * dolfin.inv(F).T * N)

    return vol_form


def get_constant(val, value_size=None, value_rank=0, constant=Constant):
    if isinstance(val, (Constant, dolfin.Constant)):
        return val
    elif isinstance(val, (Function, dolfin.Function)):
        arr = numpy_mpi.gather_vector(val.vector())
        return constant(arr)
    elif isinstance(val, dolfin.GenericVector):
        arr = numpy_mpi.gather_vector(val)
        return constant(arr)

    if value_size is None:
        if np.isscalar(val):
            value_size = 1
        else:
            try:
                value_size = len(val)
                val = np.array(val)
            except Exception as ex:
                logger.debug(ex)
                # Hope for the best
                value_size = 1
    if value_size == 1:
        if value_rank == 0:
            c = constant(val)
        else:
            c = constant([val])
    else:
        c = constant([val] * value_size)

    return c


def get_dimesion(u):
    # TODO : Check argument
    try:
        from ufl.domain import find_geometric_dimension

        dim = find_geometric_dimension(u)

    except ufl.UFLException as ex:
        try:
            dim = len(u)
        except Exception as ex2:
            logger.warning(ex)
            logger.warning(ex2)
            # Assume dimension is 3
            logger.warning("Assume dimension is 3")
            dim = 3

    return dim


def subplus(x):
    r"""
    Ramp function

    .. math::

       \max\{x,0\}

    """

    return dolfin.conditional(dolfin.ge(x, 0.0), x, 0.0)


def heaviside(x):
    r"""
    Heaviside function

    .. math::

       \frac{\mathrm{d}}{\mathrm{d}x} \max\{x,0\}

    """

    return dolfin.conditional(dolfin.ge(x, 0.0), 1.0, 0.0)


def list_sum(lst):
    """
    Return the sum of a list, when the convetiional
    method (like `sum`) it not working.
    For example if you have a list of dolfin functions.

    :param list l: a list of objects
    :returns: The sum of the list. The type depends on
              the type of elemets in the list

    """

    if not isinstance(lst, list):
        return lst

    out = lst[0]
    for item in lst[1:]:
        out += item
    return out


def QuadratureSpace(mesh, degree, dim=3):
    """
    From FEniCS version 1.6 to 2016.1 there was a change in how
    FunctionSpace is defined for quadrature spaces.
    This functions checks your dolfin version and returns the correct
    quadrature space

    :param mesh: The mesh
    :type mesh: :py:class:`dolfin.Mesh`
    :param int degree: The degree of the element
    :param int dim: For a mesh of topological dimension 3,
                    dim = 1 would be a scalar function, and
                    dim = 3 would be a vector function.
    :returns: The quadrature space
    :rtype: :py:class:`dolfin.FunctionSpace`

    """

    if DOLFIN_VERSION_MAJOR > 1.6:
        if dim == 1:
            element = dolfin.FiniteElement(
                family="Quadrature",
                cell=mesh.ufl_cell(),
                degree=degree,
                quad_scheme="default",
            )
        else:
            element = dolfin.VectorElement(
                family="Quadrature",
                cell=mesh.ufl_cell(),
                degree=degree,
                quad_scheme="default",
            )

        return dolfin.FunctionSpace(mesh, element)
    else:
        if dim == 1:
            return dolfin.FunctionSpace(mesh, "Quadrature", degree)
        else:
            return dolfin.VectorFunctionSpace(mesh, "Quadrature", degree)


class MixedParameter(Function):
    def __init__(self, fun, n, name="material_parameters"):
        """
        Initialize Mixed parameter.

        This will instanciate a function in a dolfin.MixedFunctionSpace
        consiting of `n` subspaces of the same type as `fun`.
        This is of course easy for the case when `fun` is a normal
        dolfin function, but in the case of a `RegionalParameter` it
        is not that straight forward.
        This class handles this case as well.


        :param fun: The type of you want to make a du
        :type fun: (:py:class:`dolfin.Function`)
        :param int n: number of subspaces
        :param str name: Name of the function

        .. todo::

           Implement support for MixedParameter with different
           types of subspaces, e.g [RegionalParamter, R_0, CG_1]

        """

        msg = "Please provide a dolin function as argument to MixedParameter"
        assert isinstance(fun, (dolfin.Function, Function, RegionalParameter)), msg

        if isinstance(fun, RegionalParameter):
            raise NotImplementedError

        # We can just make a usual mixed function space
        # with n copies of the original one
        V = fun.function_space()
        W = dolfin.MixedFunctionSpace([V] * n)

        Function.__init__(self, W, name=name)

        # Create a function assigner
        self.function_assigner = [FunctionAssigner(W.sub(i), V) for i in range(n)]

        # Store the original function space
        self.basespace = V

        if isinstance(fun, RegionalParameter):
            self._meshfunction = fun._meshfunction

    def assign_sub(self, f, i):
        """
        Assign subfunction

        :param f: The function you want to assign
        :param int i: The subspace number

        """
        f_ = Function(self.basespace)
        f_.assign(f)
        self.function_assigner[i].assign(self.split()[i], f_)


class RegionalParameter(Function):
    """A regional paramerter defined in terms of a dolfin.MeshFunction

    Suppose you have a MeshFunction defined different regions in your mesh,
    and you want to define different parameters on different regions,
    then this is what you want.
    """

    def __init__(self, meshfunction):
        # assert isinstance(
        #     meshfunction, dolfin.MeshFunctionSizet
        # ), "Invalid meshfunction for regional gamma"

        mesh = meshfunction.mesh()

        # FIXME
        self._values = set(numpy_mpi.gather_broadcast(meshfunction.array()))
        self._nvalues = len(self._values)

        V = dolfin.VectorFunctionSpace(mesh, "R", 0, dim=self._nvalues)

        Function.__init__(self, V)
        self._meshfunction = meshfunction

        # Functionspace for the indicator functions
        self._proj_space = dolfin.FunctionSpace(mesh, "DG", 0)

        # Make indicator functions
        self._ind_functions = []
        for v in self._values:
            self._ind_functions.append(self._make_indicator_function(v))

    def get_values(self):
        return self._values

    @property
    def function(self):
        """
        Return linear combination of coefficents
        and basis functions

        :returns: A function with parameter values at each segment
                  specified by the meshfunction
        :rtype:  :py:class`dolfin.Function

        """
        return self._sum()

    @property
    def proj_space(self):
        """
        Space for projecting the scalars.
        This is a DG 0 space.
        """
        return self._proj_space

    def _make_indicator_function(self, marker):
        dofs = self._meshfunction.where_equal(marker)
        f = Function(self._proj_space)
        f.vector()[dofs] = 1.0
        return f

    def _sum(self):
        coeffs = dolfin.split(self)
        fun = coeffs[0] * self._ind_functions[0]

        for c, f in zip(coeffs[1:], self._ind_functions[1:]):
            fun += c * f

        return fun



import os

current = os.path.dirname(os.path.abspath(__file__))


def join(p):
    return os.path.join(current, p)


mesh_paths = {
    "simple_ellipsoid": join("simple_ellipsoid.h5"),
    "prolate_ellipsoid": join("prolate_ellipsoid.h5"),
    "biv_ellipsoid": join("biv_ellipsoid.h5"),
    "benchmark": join("benchmark.h5"),
    "ellipsoid": join("ellipsoid.h5"),
}



import itertools

import dolfin
import numpy as np
import pytest

try:
    from dolfin_adjoint import (
        Constant,
        DirichletBC,
        Expression,
        UnitCubeMesh,
        interpolate,
        project,
    )
except ImportError:
    from dolfin import (
        project,
        DirichletBC,
        Constant,
        UnitCubeMesh,
        interpolate,
        Expression,
    )

from pulse import kinematics
from pulse.geometry import Geometry, MarkerFunctions, Microstructure
from pulse.material import HolzapfelOgden, NeoHookean, material_models
from pulse.mechanicsproblem import BoundaryConditions, MechanicsProblem, NeumannBC
from pulse.utils import mpi_comm_world


class Free(dolfin.SubDomain):
    def inside(self, x, on_boundary):
        return x[0] > (1.0 - dolfin.DOLFIN_EPS) and on_boundary


class Fixed(dolfin.SubDomain):
    def inside(self, x, on_boundary):
        return x[0] < dolfin.DOLFIN_EPS and on_boundary


fixed = Fixed()
fixed_marker = 1

free = Free()
free_marker = 2


@pytest.fixture
def unitcube_geometry():
    N = 3
    mesh = UnitCubeMesh(mpi_comm_world(), N, N, N)

    ffun = dolfin.MeshFunction("size_t", mesh, 2)
    ffun.set_all(0)

    fixed.mark(ffun, fixed_marker)
    free.mark(ffun, free_marker)

    marker_functions = MarkerFunctions(ffun=ffun)

    # Fibers
    V_f = dolfin.VectorFunctionSpace(mesh, "CG", 1)

    f0 = interpolate(Expression(("1.0", "0.0", "0.0"), degree=1), V_f)
    s0 = interpolate(Expression(("0.0", "1.0", "0.0"), degree=1), V_f)
    n0 = interpolate(Expression(("0.0", "0.0", "1.0"), degree=1), V_f)

    microstructure = Microstructure(f0=f0, s0=s0, n0=n0)

    geometry = Geometry(
        mesh=mesh,
        marker_functions=marker_functions,
        microstructure=microstructure,
    )

    return geometry


cases = itertools.product(
    material_models,
    ("active_strain", "active_stress"),
    (True, False),
)


@pytest.mark.parametrize("Material, active_model, isochoric", cases)
def test_material(unitcube_geometry, Material, active_model, isochoric):
    compressible_model = "incompressible"

    if active_model == "active_stress":
        active_value = 20.0
        activation = Constant(1.0)
        T_ref = active_value

        def dirichlet_bc(W):
            V = W if W.sub(0).num_sub_spaces() == 0 else W.sub(0)
            return DirichletBC(V, Constant((0.0, 0.0, 0.0)), fixed)

    else:
        activation = Constant(0.0)
        active_value = 0.0
        T_ref = 1.0

        def dirichlet_bc(W):
            V = W if W.sub(0).num_sub_spaces() == 0 else W.sub(0)
            return DirichletBC(V.sub(0), Constant(0.0), fixed, "pointwise")

    neumann_bc = NeumannBC(traction=Constant(-active_value), marker=free_marker)

    bcs = BoundaryConditions(dirichlet=(dirichlet_bc,), neumann=(neumann_bc,))

    matparams = Material.default_parameters()

    material = Material(
        activation=activation,
        parameters=matparams,
        T_ref=T_ref,
        isochoric=isochoric,
        compressible_model=compressible_model,
        active_model=active_model,
    )

    assert material.isochoric == isochoric

    problem = MechanicsProblem(unitcube_geometry, material, bcs)
    problem.solve()

    u, p = problem.state.split(deepcopy=True)

    print(material.name)
    if active_model == "active_strain":
        tol = 1e-4

        if not isochoric:
            if material.name in [
                "guccione",
                "linear_elastic",
                "saint_venant_kirchhoff",
            ]:
                assert all(abs(p.vector().get_local()) < tol)
            elif material.name == "holzapfel_ogden":
                assert all(abs(p.vector().get_local() - material.parameters["a"]) < tol)
            elif material.name == "neo_hookean":
                assert all(
                    abs(p.vector().get_local() - material.parameters["mu"]) < tol,
                )
            else:
                raise TypeError(f"Unkown material {material.name}")

        else:
            assert all(abs(p.vector().get_local()) < tol)

    else:
        F = kinematics.DeformationGradient(u)
        T = material.CauchyStress(F, p)

        V_dg = dolfin.FunctionSpace(unitcube_geometry.mesh, "DG", 1)

        # Fiber on current geometry
        f = F * unitcube_geometry.f0

        # Fiber stress
        Tf = dolfin.inner(T * f / f**2, f)
        Tf_dg = project(Tf, V_dg)

        tol = 1e-10

        assert all(abs(Tf_dg.vector().get_local() - active_value) < tol)
        assert all(abs(u.vector().get_local()) < tol)

        if not isochoric:
            if material.name in [
                "guccione",
                "linear_elastic",
                "saint_venant_kirchhoff",
            ]:
                assert all(abs(p.vector().get_local()) < tol)
            elif material.name == "holzapfel_ogden":
                assert all(abs(p.vector().get_local() - material.parameters["a"]) < tol)
            elif material.name == "neo_hookean":
                assert all(
                    abs(p.vector().get_local() - material.parameters["mu"]) < tol,
                )
            else:
                raise TypeError(f"Unkown material {material.name}")

        else:
            assert all(abs(p.vector().get_local()) < tol)


@pytest.mark.parametrize("active_model", ("active_strain", "active_stress"))
def test_active_contraction_yield_displacement(unitcube_geometry, active_model):
    activation = Constant(0.001)

    def dirichlet_bc(W):
        V = W if W.sub(0).num_sub_spaces() == 0 else W.sub(0)
        return DirichletBC(V, Constant((0.0, 0.0, 0.0)), fixed)

    bcs = BoundaryConditions(dirichlet=(dirichlet_bc,))

    matparams = HolzapfelOgden.default_parameters()

    material = HolzapfelOgden(
        activation=activation,
        parameters=matparams,
        active_model=active_model,
    )

    problem = MechanicsProblem(unitcube_geometry, material, bcs)
    problem.solve()
    u, p = problem.state.split(deepcopy=True)
    assert np.linalg.norm(u.vector().get_local()) > 0


def test_pass_active_model_as_object(unitcube_geometry):
    activation = Constant(0.001)

    def dirichlet_bc(W):
        V = W if W.sub(0).num_sub_spaces() == 0 else W.sub(0)
        return DirichletBC(V, Constant((0.0, 0.0, 0.0)), fixed)

    bcs = BoundaryConditions(dirichlet=(dirichlet_bc,))

    matparams = NeoHookean.default_parameters()

    material = NeoHookean(
        parameters=matparams,
        active_model="active_strain",
        activation=activation,
    )

    problem = MechanicsProblem(unitcube_geometry, material, bcs)
    problem.solve()
    u, p = problem.state.split(deepcopy=True)
    assert np.linalg.norm(u.vector().get_local()) > 0


if __name__ == "__main__":
    active_model = "active_stress"
    dev_iso_split = True
    geo = unitcube_geometry()
    for m in material_models:
        test_material(geo, m, active_model, dev_iso_split)



import dolfin
import numpy as np

from pulse import dolfin_utils


def test_get_constant():
    for value_size in (1, 2):
        for value_rank in (0, 1):
            vals = np.zeros(value_size)
            constant = dolfin_utils.get_constant(
                val=1,
                value_size=value_size,
                value_rank=value_rank,
            )
            constant.eval(vals, np.zeros(3))
            assert np.all((vals == 1))

            assert isinstance(constant, dolfin.Constant)


def test_map_vector_field():
    pass


def test_base_expression():
    pass


def test_regional_paramter():
    pass


def test_mixed_parameter():
    pass


if __name__ == "__main__":
    test_get_constant()



__version__ = "2023.0.0"
import logging as _logging

try:
    import dolfin as _dolfin  # noqa: F401
except ImportError:
    _logging.warning("Cannot find 'dolfin' - 'pulse' doest not work without it")



from unittest import mock

import pytest
from utils import make_mechanics_problem

from pulse import FixedPointUnloader
from pulse import HeartGeometry
from pulse import mesh_paths


@pytest.fixture
def problem():
    geo = HeartGeometry.from_file(mesh_paths["simple_ellipsoid"])
    return make_mechanics_problem(geo)


def test_fixedpointunloader(problem):
    unloader = FixedPointUnloader(
        problem=problem,
        pressure=0.1,
        options=dict(maxiter=2),
    )

    with mock.patch("pulse.mechanicsproblem.MechanicsProblem.solve") as solve_mock:
        solve_mock.return_value = (1, True)  # (niter, nconv)
        unloader.unload()


if __name__ == "__main__":
    prob = problem()
    test_fixedpointunloader(prob)



import os

import dolfin
import h5py

from .utils import getLogger
from .utils import mpi_comm_world

logger = getLogger(__name__)


parallel_h5py = h5py.h5.get_config().mpi

try:
    import mpi4py

    has_mpi4py = True
except ImportError:
    has_mpi4py = False
    if parallel_h5py:
        raise ImportError

try:
    import petsc4py

    has_petsc4py = True
except ImportError:
    has_petsc4py = False


def open_h5py(h5name, file_mode="a", comm=mpi_comm_world()):
    if parallel_h5py:
        if has_petsc4py:
            petsc4py.init()
        if has_mpi4py and has_petsc4py:
            assert isinstance(comm, (petsc4py.PETSc.Comm, mpi4py.MPI.Intracomm))

        if has_petsc4py and isinstance(comm, petsc4py.PETSc.Comm):
            comm = comm.tompi4py()

        return h5py.File(h5name, file_mode, driver="mpio", comm=comm)
    else:
        return h5py.File(h5name, file_mode)


def check_h5group(h5name, h5group, delete=False, comm=mpi_comm_world()):
    h5group_in_h5file = False
    if not os.path.isfile(h5name):
        return False

    filemode = "a" if delete else "r"
    if not os.access(h5name, os.W_OK):
        filemode = "r"
        if delete:
            logger.warning(
                ("You do not have write access to file " "{}").format(h5name),
            )
            delete = False

    with open_h5py(h5name, filemode, comm) as h5file:
        if h5group in h5file:
            h5group_in_h5file = True
            if delete:
                if parallel_h5py:
                    logger.debug(("Deleting existing group: " "'{}'").format(h5group))
                    del h5file[h5group]

                else:
                    if dolfin.MPI.rank(comm) == 0:
                        logger.debug(
                            ("Deleting existing group: " "'{}'").format(h5group),
                        )
                        del h5file[h5group]

    return h5group_in_h5file


def read_h5file(h5file, obj, group, *args, **kwargs):
    # Hack in order to work with fenics-adjoint
    # if not hasattr(obj, "create_block_variable"):
    #     obj.create_block_variable = lambda: None

    h5file.read(obj, group, *args, **kwargs)



import dolfin

from .. import kinematics
from ..dolfin_utils import get_dimesion
from ..dolfin_utils import heaviside
from ..dolfin_utils import subplus
from .active_model import ActiveModels
from .material_model import Material


class HolzapfelOgden(Material):
    r"""
    Orthotropic model by Holzapfel and Ogden

    .. math::

       \mathcal{W}(I_1, I_{4f_0})
       = \frac{a}{2 b} \left( e^{ b (I_1 - 3)}  -1 \right)
       + \frac{a_f}{2 b_f} \left( e^{ b_f (I_{4f_0} - 1)_+^2} -1 \right)
       + \frac{a_s}{2 b_s} \left( e^{ b_s (I_{4s_0} - 1)_+^2} -1 \right)
       + \frac{a_fs}{2 b_fs} \left( e^{ b_fs I_{8fs}^2} -1 \right)
    where

    .. math::

       (\cdot)_+ = \max\{x,0\}


    .. rubric:: Reference

    [1] Holzapfel, Gerhard A., and Ray W. Ogden.
    "Constitutive modelling of passive myocardium:
    a structurally based framework for material characterization.
    "Philosophical Transactions of the Royal Society of London A:
    Mathematical, Physical and Engineering Sciences 367.1902 (2009): 3445-3475.

    """
    name = "holzapfel_ogden"

    @staticmethod
    def default_parameters():
        """
        Default matereial parameter for the Holzapfel Ogden model

        Taken from Table 1 row 3 of [1]
        """

        return {
            "a": 0.059,
            "b": 8.023,
            "a_f": 18.472,
            "b_f": 16.026,
            "a_s": 2.481,
            "b_s": 11.120,
            "a_fs": 0.216,
            "b_fs": 11.436,
        }

    def W_1(self, I1, diff=0, *args, **kwargs):
        r"""
        Isotropic contribution.

        If `diff = 0`, return

        .. math::

           \frac{a}{2 b} \left( e^{ b (I_1 - 3)}  -1 \right)

        If `diff = 1`, return

        .. math::

           \frac{a}{b} e^{ b (I_1 - 3)}

        If `diff = 2`, return

        .. math::

           \frac{a b}{2}  e^{ b (I_1 - 3)}

        """

        a = self.a
        b = self.b

        if diff == 0:
            try:
                if float(a) > dolfin.DOLFIN_EPS:
                    if float(b) > dolfin.DOLFIN_EPS:
                        return a / (2.0 * b) * (dolfin.exp(b * (I1 - 3)) - 1.0)
                    else:
                        return a / 2.0 * (I1 - 3)
                else:
                    return 0.0
            except Exception:
                return a / (2.0 * b) * (dolfin.exp(b * (I1 - 3)) - 1)
        elif diff == 1:
            return a / 2.0 * dolfin.exp(b * (I1 - 3))
        elif diff == 2:
            return a * b / 2.0 * dolfin.exp(b * (I1 - 3))

    def W_4(self, I4, direction, diff=0, use_heaviside=False, *args, **kwargs):
        r"""
        Anisotropic contribution.

        If `diff = 0`, return

        .. math::

           \frac{a_f}{2 b_f} \left( e^{ b_f (I_{4f_0} - 1)_+^2} -1 \right)

        If `diff = 1`, return

        .. math::

           a_f (I_{4f_0} - 1)_+ e^{ b_f (I_{4f_0} - 1)^2}

        If `diff = 2`, return

        .. math::

           a_f h(I_{4f_0} - 1) (1 + 2b(I_{4f_0} - 1))
           e^{ b_f (I_{4f_0} - 1)_+^2}

        where

        .. math::

           h(x) = \frac{\mathrm{d}}{\mathrm{d}x} \max\{x,0\}

        is the Heaviside function.

        """
        assert direction in ["f", "s", "n"]
        a = getattr(self, f"a_{direction}")
        b = getattr(self, f"b_{direction}")

        if I4 == 0:
            return 0

        if diff == 0:
            try:
                if float(a) > dolfin.DOLFIN_EPS:
                    if float(b) > dolfin.DOLFIN_EPS:
                        return (
                            a / (2.0 * b) * (dolfin.exp(b * subplus(I4 - 1) ** 2) - 1.0)
                        )
                    else:
                        return a / 2.0 * subplus(I4 - 1) ** 2
                else:
                    return 0.0
            except Exception:
                # Probably failed to convert a and b to float
                return a / (2.0 * b) * (dolfin.exp(b * subplus(I4 - 1) ** 2) - 1.0)

        elif diff == 1:
            return a * subplus(I4 - 1) * dolfin.exp(b * pow(I4 - 1, 2))
        elif diff == 2:
            return (
                a
                * heaviside(I4 - 1)
                * (1 + 2.0 * b * pow(I4 - 1, 2))
                * dolfin.exp(b * pow(I4 - 1, 2))
            )

    def W_8(self, I8, *args, **kwargs):
        """
        Cross fiber-sheet contribution.
        """
        a = self.a_fs
        b = self.b_fs

        try:
            if float(a) > dolfin.DOLFIN_EPS:
                if float(b) > dolfin.DOLFIN_EPS:
                    return a / (2.0 * b) * (dolfin.exp(b * I8**2) - 1.0)
                else:
                    return a / 2.0 * I8**2
            else:
                return 0.0
        except Exception:
            return a / (2.0 * b) * (dolfin.exp(b * I8**2) - 1.0)

    def strain_energy(self, F):
        r"""
        Strain-energy density function.

        .. math::

           \mathcal{W} = \mathcal{W}_1 + \mathcal{W}_{4f}
           + \mathcal{W}_{\mathrm{active}}

        where

        .. math::

           \mathcal{W}_{\mathrm{active}} =
           \begin{cases}
             0 & \text{if acitve strain} \\
             \gamma I_{4f} & \text{if active stress}
           \end{cases}


        :param F: Deformation gradient
        :type F: :py:class:`dolfin.Function`

        """

        # Invariants
        I1 = kinematics.I1(F, isochoric=self.isochoric)
        I4f = kinematics.I4(F, self.f0, isochoric=self.isochoric)
        I4s = kinematics.I4(F, self.s0, isochoric=self.isochoric)
        I8fs = kinematics.I8(F, self.f0, self.s0)

        if self.active_model == ActiveModels.active_strain:
            mgamma = 1 - self.activation_field
            I1e = mgamma * I1 + (1 / mgamma**2 - mgamma) * I4f
            I4fe = 1 / mgamma**2 * I4f
            I4se = mgamma * I4s
            I8fse = 1 / dolfin.sqrt(mgamma) * I8fs
        else:
            I1e = I1
            I4fe = I4f
            I4se = I4s
            I8fse = I8fs

        # Active stress
        Wactive = self.Wactive(F, diff=0)

        dim = get_dimesion(F)
        W1 = self.W_1(I1e, diff=0, dim=dim)
        W4f = self.W_4(I4fe, "f", diff=0)
        W4s = self.W_4(I4se, "s", diff=0)
        W8fs = self.W_8(I8fse, diff=0)

        W = W1 + W4f + W4s + W8fs + Wactive

        return W



#!/usr/bin/env python
"""
This script contains the code for unloading
geometry
"""

__author__ = "Henrik Finsberg (henriknf@simula.no)"
import os

import dolfin
import numpy as np

from . import Function, interpolate

from . import unloading_utils as utils
from .mechanicsproblem import MechanicsProblem, cardiac_boundary_conditions
from .utils import getLogger, mpi_comm_world

logger = getLogger(__name__)


def setup_unloading_parameters():
    """
    Parameters for coupled unloading/material parameter
    estimation.

    For info about the different parameters,
    see the unloading module.
    """

    params = dolfin.Parameters("Unloading_parameters")

    params.add("method", "fixed_point", ["fixed_point", "raghavan"])
    # Terminate if difference in reference (unloaded) volume
    # is less than tol
    params.add("tol", 0.05)
    # Maximum number of coupled iterations
    params.add("maxiter", 5)
    # Apply conitinuation step
    params.add("continuation", False)
    # Estimate initial guess based on loaded configuration
    params.add("estimate_initial_guess", True)

    unload_options = dolfin.Parameters("unload_options")
    unload_options.add("maxiter", 10)
    unload_options.add("tol", 0.01)
    unload_options.add("ub", 2.0)
    unload_options.add("lb", 0.5)
    unload_options.add("regen_fibers", False)

    params.add(unload_options)

    return params


# try:
#     dolfin.parameters["adjoint"]["stop_annotating"] = True
# except (KeyError, RuntimeError):
#     pass


# def step(
#     geometry,
#     pressure,
#     k,
#     u,
#     residual,
#     big_res,
#     material_parameters=None,
#     params=None,
#     n=2,
#     solve_tries=1,
#     merge_control="",
#     regen_fibers=False,
# ):

#     logger.info("\n\nk = {}".format(k))

#     # Create new reference geometry by moving according to rule
#     U = Function(u.function_space())
#     numpy_mpi.assign_to_vector(U.vector(), k * numpy_mpi.gather_vector(u.vector()))

#     new_geometry = geometry.copy(u=U)

#     matparams = utils.update_material_parameters(
#         material_parameters, new_geometry.mesh, merge_control
#     )

#     if isinstance(matparams["a"], float):
#         logger.info("material parameters = {}".format(matparams["a"]))
#     else:
#         a_arr = numpy_mpi.gather_vector(matparams["a"].vector())
#         logger.info("material parameters = {}".format(a_arr))


#     # Make the solver
#     problem, active_control, passive_control = make_mechanics_problem(
#         params, new_geometry, matparams=matparams
#     )

#     try:
#         # Inflate new geometry to target pressure
#         u0 = utils.inflate_to_pressure(
#             pressure, solver, p_expr, solve_tries, n, annotate=False
#         )
#     except:
#         logger.info("Failed to increase pressure")
#         return big_res

#     # Move mesh to ED
#     mesh = Mesh(new_geometry.mesh)
#     move(mesh, u0, 1.0)

#     # Compute the residual
#     res = residual.calculate_residual(mesh)
#     logger.info("\nResidual:\t{}\n".format(res))
#     return res


class MeshUnloader(object):
    def __init__(
        self,
        problem,
        pressure,
        h5name="test.h5",
        options=None,
        h5group="",
        overwrite=False,
        merge_control="",
    ):
        self.problem = problem
        self.pressure = pressure

        self.U = Function(problem.get_displacement(annotate=True).function_space())

        self.merge_control = merge_control
        self.h5name = h5name
        self.h5group = h5group

        if os.path.isfile(h5name) and overwrite:
            if mpi_comm_world().rank == 0:
                os.remove(h5name)

        self.n = int(np.rint(np.max(np.divide(pressure, 0.4))))

        self.parameters = self.default_parameters()
        if options is not None:
            self.parameters.update(**options)

        msg = (
            "\n\n"
            + " Unloading options ".center(72, "-")
            + "\n\n"
            + f"\tTarget pressure: {pressure}\n"
            + f"\tmaxiter = {self.parameters['maxiter']}\n"
            + f"\ttolerance = {self.parameters['tol']}\n"
            + "\tregenerate_fibers (serial only)= "
            "{}\n\n".format(self.parameters["regen_fibers"]) + "".center(72, "-") + "\n"
        )
        logger.info(msg)

    def default_parameters(self):
        """
        Default parameters.
        """

        return {
            "maxiter": 10,
            "tol": 1e-4,
            "lb": 0.5,
            "ub": 2.0,
            "regen_fibers": False,
            "solve_tries": 20,
        }

    def save(self, obj, name, h5group=""):
        """
        Save object to and HDF file.

        Parameters
        ----------

        obj : dolfin.Mesh or dolfin.Function
            The object you want to save
        name : str
            Name of the object
        h5group : str
            The folder you want to save the object
            withing the HDF file. Default: ''
        """
        group = os.path.join(self.h5group, h5group)
        utils.save(obj, self.h5name, name, group)

    def unload(self, save=False):
        """
        Unload the geometry
        """
        if save:
            self.save(self.problem.geometry.mesh, "original_geometry/mesh", "0")

        logger.info("".center(72, "-"))
        logger.info("Start unloading".center(72, "-"))
        logger.info("".center(72, "-"))

        logger.info(
            (
                "\nLV Volume of original geometry = "
                "{:.3f} ml".format(self.problem.geometry.cavity_volume(chamber="lv"))
            ),
        )
        if self.problem.geometry.is_biv:
            logger.info(
                (
                    "RV Volume of original geometry = "
                    "{:.3f} ml".format(
                        self.problem.geometry.cavity_volume(chamber="rv"),
                    )
                ),
            )

        residual = utils.ResidualCalculator(self.problem.geometry.mesh)

        u = self.initial_solve(True)
        self.U = Function(u.function_space())

        self.unload_step(u, residual, save=save)

        logger.info("".center(72, "#") + "\nUnloading suceeding")

    @property
    def backward_displacement(self):
        """
        Return the current backward displacement as a
        function on the original geometry.
        """
        W = dolfin.VectorFunctionSpace(self.U.function_space().mesh(), "CG", 1)
        u_int = interpolate(self.U, W)
        u = Function(W)
        u.vector()[:] = -1 * u_int.vector()
        return u

    def initial_solve(self, save=False):
        """
        Inflate the original geometry to the target pressure and return
        the displacement field.
        """

        # Do an initial solve
        logger.info("\nDo an intial solve")
        self.problem.solve()

        u = utils.inflate_to_pressure(
            self.pressure,
            self.problem,
            self.parameters["solve_tries"],
            self.n,
            annotate=True,
        )
        return u

    @property
    def unloaded_geometry(self):
        return self.problem.geometry.copy(u=self.U, factor=-1.0)


class RaghavanUnloader(MeshUnloader):
    """
    Assumes that the given geometry is loaded
    with the given pressure.

    Find the reference configuration corresponding
    to zero pressure by first inflating the original
    geometry to the target pressure and subtacting
    the displacement times some factor `k`.
    Use a 1D minimization algorithm from scipy to find `k`.
    The method is described in [1]_
    This method assumes that material properties are given.
    Make sure to change this in utils.py

    This also runs in parallel.

    Parameters
    ----------

    geometry : object
        An obeject with attributes `mesh`, `fiber`
        and `markers`.
    pressure : float or tuple
        The target pressure. If LV only provide a
        float, if BiV proble a tuple of the form (p_LV, p_RV).
    material_parameters : dict, optional
        A dictionary with material parameters
    h5name : str, optional
        Path to where you want to save the output
        Default: test.h5
    options : dict, optional, optional
        A dictionary of solver options. All methods accept the following
        generic options:

            maxiter : int
                Maximum number of iterations to perform.
            tol : float
                Tolerance of termination
            solve_tries : int
                Number of attemtps the solver should use to
                increase the pressure (after pressure reduction)
    h5group : str
        Subfolder within the HDF file where you save the results
    remove_old : bool
        If the file with the same h5name exist, delete it before starting
    params : dict
        These are parameters that is parsed to pulse_adjoint that
        make the parameters for the solver.
        See pulse_adjoint.setup_parametere.setup_adjoint_contraction_parameters



    References
    ----------
    .. [1] Raghavan, M. L., Baoshun Ma, and Mark F. Fillinger.
       "Non-invasive determination of zero-pressure geometry of
       arterial aneurysms." Annals of biomedical engineering 34.9
       (2006): 1414-1419.

    """

    def unload_step(self, u, residual, save=True):
        raise NotImplementedError("RahavanUnloader is not yet implemented")

        # big_res = 100.0
        # residuals = {}

        # def iterate(k):

        #     res = step(
        #         self.geometry,
        #         self.pressure,
        #         k,
        #         u,
        #         residual,
        #         big_res,
        #         self.material_parameters,
        #         self.params,
        #         self.n,
        #         self.parameters["solve_tries"],
        #         self.merge_control,
        #         self.parameters["regen_fibers"],
        #     )
        #     residuals[k] = res
        #     return res

        # logger.info("\nStart iterating....")
        # res = minimize_scalar(
        #     iterate,
        #     method="bounded",
        #     bounds=(self.parameters["lb"], self.parameters["ub"]),
        #     options={
        #         "xatol": self.parameters["tol"],
        #         "maxiter": self.parameters["maxiter"],
        #     },
        # )

        # logger.info("Minimzation terminated sucessfully".center(72, "-"))
        # logger.info("Found:\n\tk={:.6f}\n\tResidual={:.3e}\n".format(res.x, res.fun))
        # logger.info("Save new reference geometry")

        # numpy_mpi.assign_to_vector(
        #     self.U.vector(), res.x * numpy_mpi.gather_vector(u.vector())
        # )
        # new_geometry = utils.update_geometry(
        #     self.geometry, self.U, self.parameters["regen_fibers"]
        # )

        # if save:
        #     self.save(new_geometry.mesh, "reference_geometry/mesh", "")


class FixedPointUnloader(MeshUnloader):
    """
    Assumes that the given geometry is loaded
    with the given pressure.

    Find the reference configuration corresponding
    to zero pressure using a backward displacement method,
    described in [2]_.
    This method assumes that material properties are given.
    Make sure to change this in utils.py

    This also runs in parallel.


    References
    ----------
    .. [2] Bols, Joris, et al. "A computational method to assess the in
        vivo stresses and unloaded configuration of patient-specific blood
        vessels." Journal of computational and Applied mathematics 246 (2013): 10-17.

    """

    def unload_step(self, u, residual, save=False, it=0):
        """
        Unload step
        """

        res = np.inf
        while it < self.parameters["maxiter"] and res > self.parameters["tol"]:
            logger.info(f"\nIteration: {it}")
            self.U.vector()[:] = u.vector()

            # The displacent field that we will move the mesh according to
            if save:
                self.save(self.U, "displacement", str(it))

            # Create new reference geomtry
            logger.debug("Create new reference geometry")
            new_geometry = self.problem.geometry.copy(u=self.U, factor=-1.0)
            utils.print_volumes(new_geometry, txt="original")
            if save:
                self.save(new_geometry.mesh, "reference_geometry/mesh", str(it))

            material = self.problem.material.copy(geometry=new_geometry)
            # TDOO: Make a copy function for bcs as well
            # use cardiac_boundary_conditions function for mechanicsproblem
            # in make_solver_paramerters
            bcs = cardiac_boundary_conditions(
                geometry=new_geometry, **self.problem.bcs_parameters
            )
            problem = MechanicsProblem(
                geometry=new_geometry,
                material=material,
                bcs=bcs,
            )

            # Solve
            u = utils.inflate_to_pressure(
                self.pressure,
                problem,
                self.parameters["solve_tries"],
                self.n,
                annotate=True,
            )

            utils.print_volumes(new_geometry, txt="inflated", u=u)

            # Move the mesh accoring to the new displacement
            ed_geometry = problem.geometry.copy(u=u, factor=1.0)
            utils.print_volumes(ed_geometry, txt="new reference")

            if save:
                self.save(ed_geometry.mesh, "ed_geometry", str(it))

            # Compute the residual
            if residual is not None:
                res = residual.calculate_residual(ed_geometry.mesh)
                logger.info(f"\nResidual:\t{res}\n")
            else:
                res = 0.0

            it += 1



import os
import sys
from pathlib import Path
from shutil import rmtree
from typing import Dict
from typing import List

from setuptools import Command
from setuptools import setup


VERSION = "2023.0.0"

here = Path(__file__).parent


# Load the package's __version__.py module as a dictionary.
about: Dict[str, str] = {}
with open(here.joinpath("pulse/__version__.py")) as f:
    exec(f.read(), about)
about["__version__"] = VERSION


class ReleaseCommand(Command):
    """Support setup.py upload."""

    description = "Build and publish the package."
    user_options: List[str] = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print(f"[1m{s}[0m")

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status("Removing previous builds…")
            rmtree(here.joinpath("dist"))
        except OSError:
            pass

        self.status("Pushing git tags…")
        os.system(f"git tag v{about['__version__']}")
        os.system("git push finsberg master --tags")

        sys.exit()


# Where the magic happens:
setup(
    version=about["__version__"],
    cmdclass={"release": ReleaseCommand},
)



from enum import Enum
from typing import Optional

import dolfin

try:
    from dolfin_adjoint import Constant
except ImportError:
    from dolfin import Constant

from ..dolfin_utils import get_dimesion, RegionalParameter


class ActiveModels(str, Enum):
    active_stress = "active_stress"
    active_strain = "active_strain"


class ActiveStressModels(str, Enum):
    transversally = "transversally"
    orthotropic = "orthotropic"
    fully_anisotropic = "fully_anisotropic"


class ActiveModel:
    def __init__(
        self,
        activation=None,
        f0=None,
        s0=None,
        n0=None,
        model: ActiveModels = ActiveModels.active_strain,
        active_isotropy: ActiveStressModels = ActiveStressModels.transversally,
        eta: Optional[float] = None,
        T_ref: Optional[float] = None,
        isochoric=True,
    ):
        # Fiber system
        self.f0 = f0
        self.s0 = s0
        self.n0 = n0

        self._activation = (
            Constant(0, name="activation") if activation is None else activation
        )

        self.T_ref = (
            Constant(T_ref, name="T_ref") if T_ref else Constant(1.0, name="T_ref")
        )

        self.eta = Constant(eta, name="eta") if eta else Constant(0.0, name="eta")
        self.isochoric = isochoric
        self.active_isotropy = active_isotropy
        self.model = model

    @property
    def dim(self):
        if not hasattr(self, "_dim"):
            try:
                self._dim = get_dimesion(self.f0)
            except Exception:
                # just assume three dimensions
                self._dim = 3
        return self._dim

    @property
    def activation(self):
        """
        Return the activation paramter.
        If regional, this will return one parameter
        for each segment.
        """
        return self._activation

    @property
    def activation_field(self):
        """
        Return the activation field.
        If regional, this will return a piecewise
        constant function (DG_0)
        """

        # Activation
        if isinstance(self._activation, RegionalParameter):
            # This means a regional activation
            # Could probably make this a bit more clean
            activation = self._activation.function
        else:
            activation = self._activation

        return self.T_ref * activation

    @property
    def Fa(self):
        if self.model == ActiveModels.active_stress:
            return dolfin.Identity(self.dim)

        f0 = self.f0
        f0f0 = dolfin.outer(f0, f0)
        Id = dolfin.Identity(self.dim)

        mgamma = 1 - self.activation_field
        Fa = mgamma * f0f0 + pow(mgamma, -1.0 / float(self.dim - 1)) * (Id - f0f0)

        return Fa

    def Fe(self, F):
        if self.model == ActiveModels.active_stress:
            return F
        Fa = self.Fa
        Fe = F * dolfin.inv(Fa)

        return Fe

    def Wactive(self, F=None, diff=0):
        """Active stress energy"""
        if self.model == ActiveModels.active_strain:
            return 0

        C = F.T * F

        if diff == 0:
            return Wactive(
                Ta=self.activation_field,
                C=C,
                f0=self.f0,
                s0=self.s0,
                n0=self.n0,
                eta=self.eta,
                active_isotropy=self.active_isotropy,
            )

        elif diff == 1:
            return self.activation_field
        raise ValueError(f"Unknown diff {diff}")


def Wactive(
    Ta,
    C,
    f0,
    eta,
    active_isotropy=ActiveStressModels.transversally,
    s0=None,
    n0=None,
):
    if active_isotropy == "transversally":
        return Wactive_transversally(
            Ta=Ta,
            C=C,
            f0=f0,
            eta=eta,
        )

    elif active_isotropy == "orthotropic":
        return Wactive_orthotropic(
            Ta=Ta,
            C=C,
            f0=f0,
            s0=s0,
            n0=n0,
        )

    elif active_isotropy == "fully_anisotropic":
        return Wactive_anisotropic(
            Ta=Ta,
            C=C,
            f0=f0,
            s0=s0,
            n0=n0,
        )
    else:
        msg = ("Unknown acitve isotropy " "{}").format(active_isotropy)
        raise ValueError(msg)


def Wactive_transversally(Ta, C, f0, eta=0.0):
    """
    Return active strain energy when activation is only
    working along the fibers, with a possible transverse
    component defined by eta

    Arguments
    ---------
    Ta : dolfin.Function or dolfin.Constant
        A scalar function representng the mangnitude of the
        active stress in the reference configuration (firt Pioala)
    C : ufl.Form
        The right Cauchy-Green deformation tensor
    f0 : dolfin.Function
        A vector function representng the direction of the
        active stress
    eta : float
        Amount of active stress in the transverse direction
        (relative to f0)
    """

    I4f = dolfin.inner(C * f0, f0)
    I1 = dolfin.tr(C)
    return Constant(0.5) * Ta * ((I4f - 1) + eta * ((I1 - 3) - (I4f - 1)))


def Wactive_orthotropic(Ta, C, f0, s0, n0):
    """Return active strain energy for an orthotropic
    active stress


    Arguments
    ---------
    Ta : dolfin.Function or dolfin.Constant
        A vector function representng the mangnitude of the
        active stress in the reference configuration (firt Pioala).
        Ta = (Ta_f0, Ta_s0, Ta_n0)
    C : ufl.Form
        The right Cauchy-Green deformation tensor
    f0 : dolfin.Function
        A vector function representng the direction of the
        first component
    s0 : dolfin.Function
        A vector function representng the direction of the
        second component
    n0 : dolfin.Function
        A vector function representng the direction of the
        third component
    """
    I4f = dolfin.inner(C * f0, f0)
    I4s = dolfin.inner(C * s0, s0)
    I4n = dolfin.inner(C * n0, n0)

    I4 = dolfin.as_vector([I4f - 1, I4s - 1, I4n - 1])
    return Constant(0.5) * dolfin.inner(Ta, I4)


def Wactive_anisotropic(Ta, C, f0, s0, n0):
    """Return active strain energy for a fully anisotropic
    acitvation.
    Note that the three basis vectors are assumed to be
    orthogonal

    Arguments
    ---------
    Ta : dolfin.Function or dolfin.Constant
        A full tensor function representng the active stress tensor
        of the active stress in the reference configuration
        (firt Pioala).
    C : ufl.Form
        The right Cauchy-Green deformation tensor
    f0 : dolfin.Function
        A vector function representng the direction of the
        first component
    s0 : dolfin.Function
        A vector function representng the direction of the
        second component
    n0 : dolfin.Function
        A vector function representng the direction of the
        third component
    """
    I4f = dolfin.inner(C * f0, f0)
    I4s = dolfin.inner(C * s0, s0)
    I4n = dolfin.inner(C * n0, n0)

    I8fs = dolfin.inner(C * f0, s0)
    I8fn = dolfin.inner(C * f0, n0)
    I8sn = dolfin.inner(C * s0, n0)

    A = dolfin.as_matrix(
        (
            (0.5 * (I4f - 1), I8fs, I8fn),
            (I8fs, 0.5 * (I4s - 1), I8sn),
            (I8fn, I8sn, 0.5 * (I4n - 1)),
        ),
    )

    return dolfin.inner(Ta, A)
